package mixer;

import java.util.ArrayList;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ArrayBlockingQueue;
import javafx.util.Pair;
import java.util.Random;
import java.util.regex.Matcher;

import com.google.protobuf.InvalidProtocolBufferException;

import meerkat.protobuf.Crypto.*;
import meerkat.protobuf.Mixing.*;

import meerkat.crypto.Encryption;
import meerkat.crypto.mixnet.Mix2ZeroKnowledgeProver;

public class Mixer implements meerkat.crypto.mixnet.Mixer{
	
	private Random random;
	private Mix2ZeroKnowledgeProver prover; 
	private Encryption encryptor;
	
	public Mixer(Random rand,Mix2ZeroKnowledgeProver prov,Encryption enc) {
		this.random = rand;
		this.prover = prov;
		this.encryptor = enc;
		
	}
	
	public Pair<ZeroKnowledgeProof[][],RerandomizableEncryptedMessage[][]> mix(List<RerandomizableEncryptedMessage> ciphertexts) throws InvalidProtocolBufferException{

		int n = ciphertexts.size();
		int nDiv2 = n >> 1;
		// assert n = 2^k and n > 1
		if( n <= 1 || ((n & (n-1)) != 0))
			return null;

		//initialization
		int layers = (int)(2 * Math.log(n)/Math.log(2)) - 1;  // layers = 2logn -1
        RerandomizableEncryptedMessage[][] encryptionTable = new RerandomizableEncryptedMessage[layers][n];
    	ZeroKnowledgeProof[][] proofsTable= new ZeroKnowledgeProof[layers][nDiv2];
		boolean[][] mixnet = createMixNet(n,layers);
        int index1, index2, switchIndex = 0;
        EncryptionRandomness r1 ,r2;
        RerandomizableEncryptedMessage e1, e2;
        boolean half = true;
        
        // set first level of encryption
        for (int j = 0; j < n; j++)
        {
            encryptionTable[0][j] = ciphertexts.get(j);
        }

        // main loop
		int i = n,iDiv2;
        for (int layer = 0; layer < layers; layer++) // i == permutation size
        {
			iDiv2 = i >> 1;
            for (int j = 0; j < n; j += i) // j == permutation index
            {
                for (int k = 0; k < iDiv2; k++) // k == elements index in permutation j
                {
                    index1 = k + j;
                    index2 = k + j + iDiv2;
                    e1 = encryptionTable[layer][index1];
                    e2 = encryptionTable[layer][index2];
                    r1 = encryptor.generateRandomness(random);
                    r2 = encryptor.generateRandomness(random);
                    if (!mixnet[layer][switchIndex])
                    {
                        encryptionTable[layer+1][index1] = encryptor.rerandomize(e1, r1);
                        encryptionTable[layer+1][index2] = encryptor.rerandomize(e2,r2);
                        
                    }
                    else
                    {
                    	encryptionTable[layer+1][index1] = encryptor.rerandomize(e2,r2);
                        encryptionTable[layer+1][index2] = encryptor.rerandomize(e1,r1);
                    }
                    proofsTable[layer][switchIndex] =
                    		prover.prove(e1, e2, encryptionTable[layer + 1][index1], 
                    				encryptionTable[layer + 1][index2],
                    				 mixnet[layer][switchIndex], r1,r2);
                    
                    switchIndex = (switchIndex + 1) % nDiv2;
                }
            }
            if (half)
            {
                i = iDiv2;
                if (i == 1)
                {
                    half = false;
                    i = 4; 			// avoid duplicate layer in the middle
                }
            }
            else
				i <<= 1;
        }
        return  new Pair<ZeroKnowledgeProof[][],RerandomizableEncryptedMessage[][]>(proofsTable, encryptionTable);
	}
	
	private int[] randomPermutation(int n){
		List<Integer> numbers= new ArrayList<Integer>(n);
		for (int i = 0; i < n; i++)
		{
			numbers.add(i);
		}

		int[] result = new int[n];
		int index;
		for (int i = 0; i < n; i++)
		{
			index = random.nextInt(n - i);
			result[i] = numbers.get(index);
			numbers.remove(index);
		}
		return result;
	}
	
	private boolean[][] createMixNet(int n,int layers)
	{
		int[] permutation = randomPermutation(n);
		int[] pi, piL, piR;
		Queue<int[]> permutationsQueue = new ArrayBlockingQueue<int[]>(n);
		Graph graph;
		int iDiv2;
		int nDiv2 = n >> 1;
		boolean[][] mixnet = new boolean[layers][nDiv2];

		permutationsQueue.add(permutation);
		for (int i = n, layer = 0; i > 1; i >>= 1, layer++) // i == permutation size
		{
			iDiv2 = i >> 1;
			for (int j = 0; j < nDiv2; j += iDiv2) // j == permutation index
			{
				pi = permutationsQueue.remove();
				graph = new Graph(pi);
				piL = new int[iDiv2];
				piR = new int[iDiv2];
				for (int k = 0; k < iDiv2; k++) // k == switch index in permutation j
				{
					mixnet[layers - layer - 1][k + j] = graph.getSwitchValue(k, true);
					mixnet[layer][k + j] = graph.getSwitchValue(k, false);

					if (!mixnet[layers - layer - 1][k + j])
					{
						piL[k] = pi[k] % iDiv2;
						piR[k] = pi[k + iDiv2] % iDiv2;
					}
					else
					{
						piL[k] = pi[k + iDiv2] % iDiv2;
						piR[k] = pi[k] % iDiv2;
					}
				}
				permutationsQueue.add(piL);
				permutationsQueue.add(piR);
			}
		}
		return mixnet;
	}

}
